<?php
/**
 * Customizer info singleton class file.
 *
 * @package WordPress
 * @subpackage Zerif Lite
 * @since Zerif Lite 1.8.5.5
 */

/**
 * Singleton class for handling the theme's customizer integration.
 *
 * @since  1.0.0
 * @access public
 */
final class Zerif_Customizer_Info_Singleton {

	/**
	 * Returns the instance.
	 *
	 * @since  1.0.0
	 * @access public
	 * @return object
	 */
	public static function get_instance() {

		static $instance = null;

		if ( is_null( $instance ) ) {
			$instance = new self;
			$instance->setup_actions();
		}

		return $instance;
	}

	/**
	 * Constructor method.
	 *
	 * @since  1.0.0
	 * @access private
	 * @return void
	 */
	private function __construct() {}

	/**
	 * Sets up initial actions.
	 *
	 * @since  1.0.0
	 * @access private
	 * @return void
	 */
	private function setup_actions() {

		// Register panels, sections, settings, controls, and partials.
		add_action( 'customize_register', array( $this, 'sections' ) );

		// Register scripts and styles for the controls.
		add_action( 'customize_controls_enqueue_scripts', array( $this, 'enqueue_control_scripts' ), 0 );
	}

	/**
	 * Sets up the customizer sections.
	 *
	 * @since  1.0.0
	 * @access public
	 * @param  object $manager WordPress customizer object.
	 * @return void
	 */
	public function sections( $manager ) {

		// Load custom sections.
		require_once( trailingslashit( get_template_directory() ) . 'inc/customizer-info/class/class-customizer-info-section.php' );

		// Register custom section types.
		$manager->register_section_type( 'Zerif_Customizer_Info' );

		// Register sections.
		$manager->add_section( new Zerif_Customizer_Info( $manager, 'zerif-upsell-big-title-section', array(
			'section_text' => sprintf( '<a href="'.esc_url( 'http://themeisle.com/themes/zerif-pro-one-page-wordpress-theme/' ).'" target="_blank">%s</a>' , __( 'View PRO version','zerif-lite' ) ).'. '.__( 'It adds a background video and a background slider.', 'zerif-lite' ),
			'panel' => 'panel_big_title',
			'priority' => 500,
		) ) );

		$manager->add_section( new Zerif_Customizer_Info( $manager, 'zerif-upsell-general-sections', array(
			'section_text' => sprintf( '<a href="'.esc_url( 'http://themeisle.com/themes/zerif-pro-one-page-wordpress-theme/' ).'" target="_blank">%s</a>' , __( 'View PRO version','zerif-lite' ) ).'. '.__( 'It adds 5 new sections, the ability to re-order existing ones and easily add custom content to frontpage.', 'zerif-lite' ),
			'panel' => 'panel_general',
			'priority' => 500,
		) ) );
	}

	/**
	 * Loads theme customizer CSS.
	 *
	 * @since  1.0.0
	 * @access public
	 * @return void
	 */
	public function enqueue_control_scripts() {
		wp_enqueue_script( 'customizer-info-js', trailingslashit( get_template_directory_uri() ) . 'inc/customizer-info/js/customizer-info-controls.js', array( 'customize-controls' ) );

		wp_enqueue_style( 'customizer-info-style', trailingslashit( get_template_directory_uri() ) . 'inc/customizer-info/css/customizer-info.css' );
	}
}

Zerif_Customizer_Info_Singleton::get_instance();
